@@ -1,7 +1,8 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, MariaDB Corporation. All Rights Reserved.
+Copyright (c) 2013, 2017, MariaDB Corporation.
+Copyright (c) 2013, 2014, Fusion-io
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -31,6 +32,7 @@ Created 11/11/1995 Heikki Tuuri
 #endif
 
 #include "buf0buf.h"
+#include "buf0mtflu.h"
 #include "buf0checksum.h"
 #include "srv0start.h"
 #include "srv0srv.h"
@@ -45,10 +47,12 @@ Created 11/11/1995 Heikki Tuuri
 #include "ibuf0ibuf.h"
 #include "log0log.h"
 #include "os0file.h"
+#include "os0sync.h"
 #include "trx0sys.h"
 #include "srv0mon.h"
 #include "mysql/plugin.h"
 #include "mysql/service_thd_wait.h"
+#include "fil0pagecompress.h"
 
 /** Number of pages flushed through non flush_list flushes. */
 static ulint buf_lru_flush_page_count = 0;
@@ -60,14 +64,13 @@ need to protect it by a mutex. It is only ever read by the thread
 doing the shutdown */
 UNIV_INTERN ibool buf_page_cleaner_is_active = FALSE;
 
-/** LRU flush batch is further divided into this chunk size to
-reduce the wait time for the threads waiting for a clean block */
-#define PAGE_CLEANER_LRU_BATCH_CHUNK_SIZE	100
-
 #ifdef UNIV_PFS_THREAD
 UNIV_INTERN mysql_pfs_key_t buf_page_cleaner_thread_key;
 #endif /* UNIV_PFS_THREAD */
 
+/** Event to synchronise with the flushing. */
+ os_event_t	buf_flush_event;
+
 /** If LRU list of a buf_pool is less than this size then LRU eviction
 should not happen. This is because when we do LRU flushing we also put
 the blocks on free list. If LRU list is very small then we can end up
@@ -76,15 +79,6 @@ in thrashing. */
 
 /* @} */
 
-/** Handled page counters for a single flush */
-struct flush_counters_t {
-	ulint	flushed;	/*!< number of dirty pages flushed */
-	ulint	evicted;	/*!< number of clean pages evicted, including
-			        evicted uncompressed page images */
-	ulint	unzip_LRU_evicted;/*!< number of uncompressed page images
-				evicted */
-};
-
 /******************************************************************//**
 Increases flush_list size in bytes with zip_size for compressed page,
 UNIV_PAGE_SIZE for uncompressed page in inline function */
@@ -140,60 +134,6 @@ buf_flush_validate_skip(
 }
 #endif /* UNIV_DEBUG || UNIV_BUF_DEBUG */
 
-/*******************************************************************//**
-Sets hazard pointer during flush_list iteration. */
-UNIV_INLINE
-void
-buf_flush_set_hp(
-/*=============*/
-	buf_pool_t*		buf_pool,/*!< in/out: buffer pool instance */
-	const buf_page_t*	bpage)	/*!< in: buffer control block */
-{
-	ut_ad(buf_flush_list_mutex_own(buf_pool));
-	ut_ad(buf_pool->flush_list_hp == NULL || bpage == NULL);
-	ut_ad(!bpage || buf_page_in_file(bpage));
-	ut_ad(!bpage || bpage->in_flush_list);
-	ut_ad(!bpage || buf_pool_from_bpage(bpage) == buf_pool);
-
-	buf_pool->flush_list_hp = bpage;
-}
-
-/*******************************************************************//**
-Checks if the given block is a hazard pointer
-@return true if bpage is hazard pointer */
-UNIV_INLINE
-bool
-buf_flush_is_hp(
-/*============*/
-	buf_pool_t*		buf_pool,/*!< in: buffer pool instance */
-	const buf_page_t*	bpage)	/*!< in: buffer control block */
-{
-	ut_ad(buf_flush_list_mutex_own(buf_pool));
-
-	return(buf_pool->flush_list_hp == bpage);
-}
-
-/*******************************************************************//**
-Whenever we move a block in flush_list (either to remove it or to
-relocate it) we check the hazard pointer set by some other thread
-doing the flush list scan. If the hazard pointer is the same as the
-one we are about going to move then we set it to NULL to force a rescan
-in the thread doing the batch. */
-UNIV_INLINE
-void
-buf_flush_update_hp(
-/*================*/
-	buf_pool_t*	buf_pool,	/*!< in: buffer pool instance */
-	buf_page_t*	bpage)		/*!< in: buffer control block */
-{
-	ut_ad(buf_flush_list_mutex_own(buf_pool));
-
-	if (buf_flush_is_hp(buf_pool, bpage)) {
-		buf_flush_set_hp(buf_pool, NULL);
-		MONITOR_INC(MONITOR_FLUSH_HP_RESCAN);
-	}
-}
-
 /******************************************************************//**
 Insert a block in the flush_rbt and returns a pointer to its
 predecessor or NULL if no predecessor. The ordering is maintained
@@ -210,6 +150,7 @@ buf_flush_insert_in_flush_rbt(
 	buf_page_t*		prev = NULL;
 	buf_pool_t*		buf_pool = buf_pool_from_bpage(bpage);
 
+	ut_ad(srv_shutdown_state != SRV_SHUTDOWN_FLUSH_PHASE);
 	ut_ad(buf_flush_list_mutex_own(buf_pool));
 
 	/* Insert this buffer into the rbt. */
@@ -419,6 +360,7 @@ buf_flush_insert_sorted_into_flush_list(
 	buf_page_t*	prev_b;
 	buf_page_t*	b;
 
+	ut_ad(srv_shutdown_state != SRV_SHUTDOWN_FLUSH_PHASE);
 	ut_ad(!buf_pool_mutex_own(buf_pool));
 	ut_ad(log_flush_order_mutex_own());
 	ut_ad(mutex_own(&block->mutex));
@@ -592,6 +534,10 @@ buf_flush_remove(
 
 	buf_flush_list_mutex_enter(buf_pool);
 
+	/* Important that we adjust the hazard pointer before removing
+	the bpage from flush list. */
+	buf_pool->flush_hp.adjust(bpage);
+
 	switch (buf_page_get_state(bpage)) {
 	case BUF_BLOCK_POOL_WATCH:
 	case BUF_BLOCK_ZIP_PAGE:
@@ -632,7 +578,6 @@ buf_flush_remove(
 	ut_a(buf_flush_validate_skip(buf_pool));
 #endif /* UNIV_DEBUG || UNIV_BUF_DEBUG */
 
-	buf_flush_update_hp(buf_pool, bpage);
 	buf_flush_list_mutex_exit(buf_pool);
 }
 
@@ -683,6 +628,10 @@ buf_flush_relocate_on_flush_list(
 		prev_b = buf_flush_insert_in_flush_rbt(dpage);
 	}
 
+	/* Important that we adjust the hazard pointer before removing
+	the bpage from the flush list. */
+	buf_pool->flush_hp.adjust(bpage);
+
 	/* Must be done after we have removed it from the flush_rbt
 	because we assert on in_flush_list in comparison function. */
 	ut_d(bpage->in_flush_list = FALSE);
@@ -711,7 +660,6 @@ buf_flush_relocate_on_flush_list(
 	ut_a(buf_flush_validate_low(buf_pool));
 #endif /* UNIV_DEBUG || UNIV_BUF_DEBUG */
 
-	buf_flush_update_hp(buf_pool, bpage);
 	buf_flush_list_mutex_exit(buf_pool);
 }
 
@@ -732,6 +680,7 @@ buf_flush_write_complete(
 
 	flush_type = buf_page_get_flush_type(bpage);
 	buf_pool->n_flush[flush_type]--;
+	ut_ad(buf_pool->n_flush[flush_type] != ULINT_MAX);
 
 	ut_ad(buf_pool_mutex_own(buf_pool));
 
@@ -766,7 +715,7 @@ buf_flush_update_zip_checksum(
 				srv_checksum_algorithm)));
 
 	mach_write_to_8(page + FIL_PAGE_LSN, lsn);
-	memset(page + FIL_PAGE_FILE_FLUSH_LSN, 0, 8);
+	memset(page + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, 0, 8);
 	mach_write_to_4(page + FIL_PAGE_SPACE_OR_CHKSUM, checksum);
 }
 
@@ -885,9 +834,12 @@ buf_flush_write_block_low(
 	buf_flush_t	flush_type,	/*!< in: type of flush */
 	bool		sync)		/*!< in: true if sync IO request */
 {
+	fil_space_t*	space = fil_space_acquire_for_io(bpage->space);
+	if (!space) {
+		return;
+	}
 	ulint	zip_size	= buf_page_get_zip_size(bpage);
 	page_t*	frame		= NULL;
-
 #ifdef UNIV_DEBUG
 	buf_pool_t*	buf_pool = buf_pool_from_bpage(bpage);
 	ut_ad(!buf_pool_mutex_own(buf_pool));
@@ -943,7 +895,7 @@ buf_flush_write_block_low(
 
 		ut_a(page_zip_verify_checksum(frame, zip_size));
 
-		memset(frame + FIL_PAGE_FILE_FLUSH_LSN, 0, 8);
+		memset(frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, 0, 8);
 		break;
 	case BUF_BLOCK_FILE_PAGE:
 		frame = bpage->zip.data;
@@ -958,17 +910,46 @@ buf_flush_write_block_low(
 		break;
 	}
 
+	frame = buf_page_encrypt_before_write(space, bpage, frame);
+
 	if (!srv_use_doublewrite_buf || !buf_dblwr) {
 		fil_io(OS_FILE_WRITE | OS_AIO_SIMULATED_WAKE_LATER,
-		       sync, buf_page_get_space(bpage), zip_size,
-		       buf_page_get_page_no(bpage), 0,
-		       zip_size ? zip_size : UNIV_PAGE_SIZE,
-		       frame, bpage);
-	} else if (flush_type == BUF_FLUSH_SINGLE_PAGE) {
-		buf_dblwr_write_single_page(bpage, sync);
+			sync,
+			buf_page_get_space(bpage),
+			zip_size,
+			buf_page_get_page_no(bpage),
+			0,
+			zip_size ? zip_size : bpage->real_size,
+			frame,
+			bpage,
+			&bpage->write_size);
 	} else {
-		ut_ad(!sync);
-		buf_dblwr_add_to_batch(bpage);
+
+		/* InnoDB uses doublewrite buffer and doublewrite buffer
+		is initialized. User can define do we use atomic writes
+		on a file space (table) or not. If atomic writes are
+		not used we should use doublewrite buffer and if
+		atomic writes should be used, no doublewrite buffer
+		is used. */
+
+		if (fsp_flags_get_atomic_writes(space->flags)
+		    == ATOMIC_WRITES_ON) {
+			fil_io(OS_FILE_WRITE | OS_AIO_SIMULATED_WAKE_LATER,
+				FALSE,
+				buf_page_get_space(bpage),
+				zip_size,
+				buf_page_get_page_no(bpage),
+				0,
+				zip_size ? zip_size : bpage->real_size,
+				frame,
+				bpage,
+				&bpage->write_size);
+		} else if (flush_type == BUF_FLUSH_SINGLE_PAGE) {
+			buf_dblwr_write_single_page(bpage, sync);
+		} else {
+			ut_ad(!sync);
+			buf_dblwr_add_to_batch(bpage);
+		}
 	}
 
 	/* When doing single page flushing the IO is done synchronously
@@ -976,10 +957,27 @@ buf_flush_write_block_low(
 	are working on. */
 	if (sync) {
 		ut_ad(flush_type == BUF_FLUSH_SINGLE_PAGE);
-		fil_flush(buf_page_get_space(bpage));
-		buf_page_io_complete(bpage);
+		fil_flush(space);
+
+		/* The tablespace could already have been dropped,
+		because fil_io(request, sync) would already have
+		decremented the node->n_pending. However,
+		buf_page_io_complete() only needs to look up the
+		tablespace during read requests, not during writes. */
+		ut_ad(buf_page_get_io_fix(bpage) == BUF_IO_WRITE);
+
+		/* true means we want to evict this page from the
+		LRU list as well. */
+#ifdef UNIV_DEBUG
+		dberr_t err =
+#endif
+		buf_page_io_complete(bpage, true);
+
+		ut_ad(err == DB_SUCCESS);
 	}
 
+	fil_space_release_for_io(space);
+
 	/* Increment the counter of I/O operations used
 	for selecting LRU policy. */
 	buf_LRU_stat_inc_io();
@@ -1056,6 +1054,7 @@ buf_flush_page(
 		}
 
 		++buf_pool->n_flush[flush_type];
+		ut_ad(buf_pool->n_flush[flush_type] != 0);
 
 		mutex_exit(block_mutex);
 		buf_pool_mutex_exit(buf_pool);
@@ -1071,10 +1070,10 @@ buf_flush_page(
 			rw_lock_s_lock_gen(rw_lock, BUF_IO_WRITE);
                 }
 
-                /* Even though bpage is not protected by any mutex at this
-                point, it is safe to access bpage, because it is io_fixed and
-                oldest_modification != 0.  Thus, it cannot be relocated in the
-                buffer pool or removed from flush_list or LRU_list. */
+		/* Even though bpage is not protected by any mutex at this
+		point, it is safe to access bpage, because it is io_fixed and
+		oldest_modification != 0.  Thus, it cannot be relocated in the
+		buffer pool or removed from flush_list or LRU_list. */
 
                 buf_flush_write_block_low(bpage, flush_type, sync);
         }
@@ -1228,7 +1227,9 @@ buf_flush_try_neighbors(
 		}
 	}
 
+#ifdef UNIV_DEBUG
 	/* fprintf(stderr, "Flush area: low %lu high %lu\n", low, high); */
+#endif
 
 	if (high > fil_space_get_size(space)) {
 		high = fil_space_get_size(space);
@@ -1436,9 +1437,8 @@ This utility flushes dirty blocks from the end of the LRU list.
 The calling thread is not allowed to own any latches on pages!
 It attempts to make 'max' blocks available in the free list. Note that
 it is a best effort attempt and it is not guaranteed that after a call
-to this function there will be 'max' blocks in the free list.
-@return number of blocks for which the write request was queued. */
-static
+to this function there will be 'max' blocks in the free list.*/
+__attribute__((nonnull))
 void
 buf_flush_LRU_list_batch(
 /*=====================*/
@@ -1449,96 +1449,54 @@ buf_flush_LRU_list_batch(
 					counts */
 {
 	buf_page_t*	bpage;
-	ulint		count = 0;
 	ulint		scanned = 0;
 	ulint		free_len = UT_LIST_GET_LEN(buf_pool->free);
 	ulint		lru_len = UT_LIST_GET_LEN(buf_pool->LRU);
 
-	ut_ad(buf_pool_mutex_own(buf_pool));
-
 	n->flushed = 0;
 	n->evicted = 0;
 	n->unzip_LRU_evicted = 0;
 
-	bpage = UT_LIST_GET_LAST(buf_pool->LRU);
-	while (bpage != NULL && count < max
-	       && (n->flushed + n->evicted) < max
-	       && free_len < srv_LRU_scan_depth
-	       && lru_len > BUF_LRU_MIN_LEN) {
+	ut_ad(buf_pool_mutex_own(buf_pool));
+
+	for (bpage = UT_LIST_GET_LAST(buf_pool->LRU);
+	     bpage != NULL && (n->evicted + n->flushed) < max
+	     && free_len < srv_LRU_scan_depth
+	     && lru_len > BUF_LRU_MIN_LEN;
+	     ++scanned,
+	     bpage = buf_pool->lru_hp.get()) {
+
+		buf_page_t* prev = UT_LIST_GET_PREV(LRU, bpage);
+		buf_pool->lru_hp.set(prev);
 
 		ib_mutex_t* block_mutex = buf_page_get_mutex(bpage);
-		ibool	 evict;
-
 		mutex_enter(block_mutex);
-		evict = buf_flush_ready_for_replace(bpage);
+		bool	evict = buf_flush_ready_for_replace(bpage);
 		mutex_exit(block_mutex);
 
-		++scanned;
-
-		/* If the block is ready to be replaced we try to
-		free it i.e.: put it on the free list.
-		Otherwise we try to flush the block and its
-		neighbors. In this case we'll put it on the
-		free list in the next pass. We do this extra work
-		of putting blocks to the free list instead of
-		just flushing them because after every flush
-		we have to restart the scan from the tail of
-		the LRU list and if we don't clear the tail
-		of the flushed pages then the scan becomes
-		O(n*n). */
 		if (evict) {
+			/* block is ready for eviction i.e., it is
+			clean and is not IO-fixed or buffer fixed. */
 			if (buf_LRU_free_page(bpage, true)) {
-				/* buf_pool->mutex was potentially
-				released and reacquired. */
 				n->evicted++;
-				bpage = UT_LIST_GET_LAST(buf_pool->LRU);
-			} else {
-				bpage = UT_LIST_GET_PREV(LRU, bpage);
 			}
 		} else {
-			ulint		space;
-			ulint		offset;
-			buf_page_t*	prev_bpage;
-
-			prev_bpage = UT_LIST_GET_PREV(LRU, bpage);
-
-			/* Save the previous bpage */
-
-			if (prev_bpage != NULL) {
-				space = prev_bpage->space;
-				offset = prev_bpage->offset;
-			} else {
-				space = ULINT_UNDEFINED;
-				offset = ULINT_UNDEFINED;
-			}
-
-			if (!buf_flush_page_and_try_neighbors(
-				bpage, BUF_FLUSH_LRU, max, &n->flushed)) {
-
-				bpage = prev_bpage;
-			} else {
-				/* buf_pool->mutex was released.
-				reposition the iterator. Note: the
-				prev block could have been repositioned
-				too but that should be rare. */
-
-				if (prev_bpage != NULL) {
-
-					ut_ad(space != ULINT_UNDEFINED);
-					ut_ad(offset != ULINT_UNDEFINED);
-
-					prev_bpage = buf_page_hash_get(
-						buf_pool, space, offset);
-				}
-
-				bpage = prev_bpage;
-			}
+			/* Block is ready for flush. Dispatch an IO
+			request. The IO helper thread will put it on
+			free list in IO completion routine. */
+			buf_flush_page_and_try_neighbors(
+				bpage, BUF_FLUSH_LRU, max, &n->flushed);
 		}
 
+		ut_ad(!mutex_own(block_mutex));
+		ut_ad(buf_pool_mutex_own(buf_pool));
+
 		free_len = UT_LIST_GET_LEN(buf_pool->free);
 		lru_len = UT_LIST_GET_LEN(buf_pool->LRU);
 	}
 
+	buf_pool->lru_hp.set(NULL);
+
 	/* We keep track of all flushes happening as part of LRU
 	flush. When estimating the desired rate at which flush_list
 	should be flushed, we factor in this value. */
@@ -1557,10 +1515,8 @@ buf_flush_LRU_list_batch(
 
 /*******************************************************************//**
 Flush and move pages from LRU or unzip_LRU list to the free list.
-Whether LRU or unzip_LRU is used depends on the state of the system.
-@return number of blocks for which either the write request was queued
-or in case of unzip_LRU the number of blocks actually moved to the
-free list */
+Whether LRU or unzip_LRU is used depends on the state of the system.*/
+__attribute__((nonnull))
 static
 void
 buf_do_LRU_batch(
@@ -1571,7 +1527,6 @@ buf_do_LRU_batch(
 	flush_counters_t*	n)	/*!< out: flushed/evicted page
 					counts */
 {
-
 	if (buf_LRU_evict_from_unzip_LRU(buf_pool)) {
 		n->unzip_LRU_evicted = buf_free_from_unzip_LRU_list_batch(buf_pool, max);
 	} else {
@@ -1584,6 +1539,10 @@ buf_do_LRU_batch(
 		n->evicted = 0;
 		n->flushed = 0;
 	}
+
+	/* Add evicted pages from unzip_LRU to the evicted pages from
+	the simple LRU. */
+	n->evicted += n->unzip_LRU_evicted;
 }
 
 /*******************************************************************//**
@@ -1625,6 +1584,7 @@ buf_do_flush_list_batch(
 	for (buf_page_t* bpage = UT_LIST_GET_LAST(buf_pool->flush_list);
 	     count < min_n && bpage != NULL && len > 0
 	     && bpage->oldest_modification < lsn_limit;
+	     bpage = buf_pool->flush_hp.get(),
 	     ++scanned) {
 
 		buf_page_t*	prev;
@@ -1633,8 +1593,7 @@ buf_do_flush_list_batch(
 		ut_ad(bpage->in_flush_list);
 
 		prev = UT_LIST_GET_PREV(list, bpage);
-		buf_flush_set_hp(buf_pool, prev);
-
+		buf_pool->flush_hp.set(prev);
 		buf_flush_list_mutex_exit(buf_pool);
 
 #ifdef UNIV_DEBUG
@@ -1645,23 +1604,12 @@ buf_do_flush_list_batch(
 
 		buf_flush_list_mutex_enter(buf_pool);
 
-		ut_ad(flushed || buf_flush_is_hp(buf_pool, prev));
+		ut_ad(flushed || buf_pool->flush_hp.is_hp(prev));
 
-		if (!buf_flush_is_hp(buf_pool, prev)) {
-			/* The hazard pointer was reset by some other
-			thread. Restart the scan. */
-			ut_ad(buf_flush_is_hp(buf_pool, NULL));
-			bpage = UT_LIST_GET_LAST(buf_pool->flush_list);
-			len = UT_LIST_GET_LEN(buf_pool->flush_list);
-		} else {
-			bpage = prev;
-			--len;
-			buf_flush_set_hp(buf_pool, NULL);
-		}
-
-		ut_ad(!bpage || bpage->in_flush_list);
+		--len;
 	}
 
+	buf_pool->flush_hp.set(NULL);
 	buf_flush_list_mutex_exit(buf_pool);
 
 	MONITOR_INC_VALUE_CUMULATIVE(MONITOR_FLUSH_BATCH_SCANNED,
@@ -1679,9 +1627,8 @@ This utility flushes dirty blocks from the end of the LRU list or flush_list.
 NOTE 1: in the case of an LRU flush the calling thread may own latches to
 pages: to avoid deadlocks, this function must be written so that it cannot
 end up waiting for these latches! NOTE 2: in the case of a flush list flush,
-the calling thread is not allowed to own any latches on pages!
-@return number of blocks for which the write request was queued */
-static
+the calling thread is not allowed to own any latches on pages! */
+__attribute__((nonnull))
 void
 buf_flush_batch(
 /*============*/
@@ -1701,7 +1648,6 @@ buf_flush_batch(
 	flush_counters_t*	n)	/*!< out: flushed/evicted page
 					counts  */
 {
-
 	ut_ad(flush_type == BUF_FLUSH_LRU || flush_type == BUF_FLUSH_LIST);
 #ifdef UNIV_SYNC_DEBUG
 	ut_ad((flush_type != BUF_FLUSH_LIST)
@@ -1738,7 +1684,6 @@ buf_flush_batch(
 
 /******************************************************************//**
 Gather the aggregated stats for both flush list and LRU list flushing */
-static
 void
 buf_flush_common(
 /*=============*/
@@ -1763,7 +1708,6 @@ buf_flush_common(
 
 /******************************************************************//**
 Start a buffer flush batch for LRU or flush list */
-static
 ibool
 buf_flush_start(
 /*============*/
@@ -1792,7 +1736,6 @@ buf_flush_start(
 
 /******************************************************************//**
 End a buffer flush batch for LRU or flush list */
-static
 void
 buf_flush_end(
 /*==========*/
@@ -1847,40 +1790,6 @@ buf_flush_wait_batch_end(
 	}
 }
 
-/*******************************************************************//**
-This utility flushes dirty blocks from the end of the LRU list and also
-puts replaceable clean pages from the end of the LRU list to the free
-list.
-NOTE: The calling thread is not allowed to own any latches on pages!
-@return true if a batch was queued successfully. false if another batch
-of same type was already running. */
-static
-bool
-buf_flush_LRU(
-/*==========*/
-	buf_pool_t*	buf_pool,	/*!< in/out: buffer pool instance */
-	ulint		min_n,		/*!< in: wished minimum mumber of blocks
-					flushed (it is not guaranteed that the
-					actual number is that big, though) */
-	flush_counters_t	*n)	/*!< out: flushed/evicted page
-					counts */
-{
-	if (!buf_flush_start(buf_pool, BUF_FLUSH_LRU)) {
-		n->flushed = 0;
-		n->evicted = 0;
-		n->unzip_LRU_evicted = 0;
-		return(false);
-	}
-
-	buf_flush_batch(buf_pool, BUF_FLUSH_LRU, min_n, 0, n);
-
-	buf_flush_end(buf_pool, BUF_FLUSH_LRU);
-
-	buf_flush_common(BUF_FLUSH_LRU, n->flushed);
-
-	return(true);
-}
-
 /*******************************************************************//**
 This utility flushes dirty blocks from the end of the flush list of
 all buffer pool instances.
@@ -1908,6 +1817,10 @@ buf_flush_list(
 	ulint		i;
 	bool		success = true;
 
+	if (buf_mtflu_init_done()) {
+		return(buf_mtflu_flush_list(min_n, lsn_limit, n_processed));
+	}
+
 	if (n_processed) {
 		*n_processed = 0;
 	}
@@ -1923,8 +1836,8 @@ buf_flush_list(
 
 	/* Flush to lsn_limit in all buffer pool instances */
 	for (i = 0; i < srv_buf_pool_instances; i++) {
-		buf_pool_t*	buf_pool;
-		flush_counters_t n;
+		buf_pool_t*		buf_pool;
+		flush_counters_t	n;
 
 		buf_pool = buf_pool_from_array(i);
 
@@ -1968,12 +1881,12 @@ buf_flush_list(
 }
 
 /******************************************************************//**
-This function picks up a single dirty page from the tail of the LRU
-list, flushes it, removes it from page_hash and LRU list and puts
-it on the free list. It is called from user threads when they are
-unable to find a replaceable page at the tail of the LRU list i.e.:
-when the background LRU flushing in the page_cleaner thread is not
-fast enough to keep pace with the workload.
+This function picks up a single page from the tail of the LRU
+list, flushes it (if it is dirty), removes it from page_hash and LRU
+list and puts it on the free list. It is called from user threads when
+they are unable to find a replaceable page at the tail of the LRU
+list i.e.: when the background LRU flushing in the page_cleaner thread
+is not fast enough to keep pace with the workload.
 @return TRUE if success. */
 UNIV_INTERN
 ibool
@@ -1983,84 +1896,67 @@ buf_flush_single_page_from_LRU(
 {
 	ulint		scanned;
 	buf_page_t*	bpage;
+	ibool		freed;
 
 	buf_pool_mutex_enter(buf_pool);
 
-	for (bpage = UT_LIST_GET_LAST(buf_pool->LRU), scanned = 1;
+	for (bpage = buf_pool->single_scan_itr.start(),
+	     scanned = 0, freed = FALSE;
 	     bpage != NULL;
-	     bpage = UT_LIST_GET_PREV(LRU, bpage), ++scanned) {
+	     ++scanned, bpage = buf_pool->single_scan_itr.get()) {
 
-		ib_mutex_t*	block_mutex = buf_page_get_mutex(bpage);
+		ut_ad(buf_pool_mutex_own(buf_pool));
 
+		buf_page_t* prev = UT_LIST_GET_PREV(LRU, bpage);
+		buf_pool->single_scan_itr.set(prev);
+
+		ib_mutex_t* block_mutex = buf_page_get_mutex(bpage);
 		mutex_enter(block_mutex);
 
-		if (buf_flush_ready_for_flush(bpage, BUF_FLUSH_SINGLE_PAGE)) {
-
-			/* The following call will release the buffer pool
-			and block mutex. */
-
-			ibool	flushed = buf_flush_page(
-				buf_pool, bpage, BUF_FLUSH_SINGLE_PAGE, true);
-
-			if (flushed) {
-				/* buf_flush_page() will release the
-				block mutex */
+		if (buf_flush_ready_for_replace(bpage)) {
+			/* block is ready for eviction i.e., it is
+			clean and is not IO-fixed or buffer fixed. */
+			mutex_exit(block_mutex);
+			if (buf_LRU_free_page(bpage, true)) {
+				buf_pool_mutex_exit(buf_pool);
+				freed = TRUE;
 				break;
 			}
+		} else if (buf_flush_ready_for_flush(
+				bpage, BUF_FLUSH_SINGLE_PAGE)) {
+			/* Block is ready for flush. Dispatch an IO
+			request. We'll put it on free list in IO
+			completion routine. The following call, if
+			successful, will release the buffer pool and
+			block mutex. */
+			freed = buf_flush_page(buf_pool, bpage,
+					       BUF_FLUSH_SINGLE_PAGE, true);
+			if (freed) {
+				/* block and buffer pool mutex have
+				already been reelased. */
+				break;
+			}
+			mutex_exit(block_mutex);
+		} else {
+			mutex_exit(block_mutex);
 		}
-
-		mutex_exit(block_mutex);
 	}
 
-	MONITOR_INC_VALUE_CUMULATIVE(
-		MONITOR_LRU_SINGLE_FLUSH_SCANNED,
-		MONITOR_LRU_SINGLE_FLUSH_SCANNED_NUM_CALL,
-		MONITOR_LRU_SINGLE_FLUSH_SCANNED_PER_CALL,
-		scanned);
-
-	if (bpage == NULL) {
+	if (!freed) {
 		/* Can't find a single flushable page. */
+		ut_ad(!bpage);
 		buf_pool_mutex_exit(buf_pool);
-		return(FALSE);
 	}
 
-
-	ibool	freed = FALSE;
-
-	/* At this point the page has been written to the disk.
-	As we are not holding buffer pool or block mutex therefore
-	we cannot use the bpage safely. It may have been plucked out
-	of the LRU list by some other thread or it may even have
-	relocated in case of a compressed page. We need to start
-	the scan of LRU list again to remove the block from the LRU
-	list and put it on the free list. */
-	buf_pool_mutex_enter(buf_pool);
-
-	for (bpage = UT_LIST_GET_LAST(buf_pool->LRU);
-	     bpage != NULL;
-	     bpage = UT_LIST_GET_PREV(LRU, bpage)) {
-
-		ib_mutex_t*	block_mutex = buf_page_get_mutex(bpage);
-
-		mutex_enter(block_mutex);
-
-		ibool	ready = buf_flush_ready_for_replace(bpage);
-
-		mutex_exit(block_mutex);
-
-		if (ready) {
-			bool	evict_zip;
-
-			evict_zip = !buf_LRU_evict_from_unzip_LRU(buf_pool);;
-
-			freed = buf_LRU_free_page(bpage, evict_zip);
-
-			break;
-		}
+	if (scanned) {
+		MONITOR_INC_VALUE_CUMULATIVE(
+			MONITOR_LRU_SINGLE_FLUSH_SCANNED,
+			MONITOR_LRU_SINGLE_FLUSH_SCANNED_NUM_CALL,
+			MONITOR_LRU_SINGLE_FLUSH_SCANNED_PER_CALL,
+			scanned);
 	}
 
-	buf_pool_mutex_exit(buf_pool);
-
+	ut_ad(!buf_pool_mutex_own(buf_pool));
 	return(freed);
 }
 
@@ -2078,10 +1974,16 @@ buf_flush_LRU_tail(void)
 {
 	ulint	total_flushed = 0;
 
+	if(buf_mtflu_init_done())
+	{
+		return(buf_mtflu_flush_LRU_tail());
+	}
+
 	for (ulint i = 0; i < srv_buf_pool_instances; i++) {
 
 		buf_pool_t*	buf_pool = buf_pool_from_array(i);
 		ulint		scan_depth;
+		flush_counters_t	n;
 
 		/* srv_LRU_scan_depth can be arbitrarily large value.
 		We cap it with current LRU size. */
@@ -2091,44 +1993,37 @@ buf_flush_LRU_tail(void)
 
 		scan_depth = ut_min(srv_LRU_scan_depth, scan_depth);
 
-		/* We divide LRU flush into smaller chunks because
-		there may be user threads waiting for the flush to
-		end in buf_LRU_get_free_block(). */
-		for (ulint j = 0;
-		     j < scan_depth;
-		     j += PAGE_CLEANER_LRU_BATCH_CHUNK_SIZE) {
-
-			flush_counters_t	n;
-
-			/* Currently page_cleaner is the only thread
-			that can trigger an LRU flush. It is possible
-			that a batch triggered during last iteration is
-			still running, */
-			if (buf_flush_LRU(buf_pool,
-					  PAGE_CLEANER_LRU_BATCH_CHUNK_SIZE,
-					  &n)) {
-
-				/* Allowed only one batch per
-				buffer pool instance. */
-				buf_flush_wait_batch_end(
-					buf_pool, BUF_FLUSH_LRU);
-			}
-
-			if (n.flushed) {
-				total_flushed += n.flushed;
-			} else {
-				/* Nothing to flush */
-				break;
-			}
+		/* Currently page_cleaner is the only thread
+		that can trigger an LRU flush. It is possible
+		that a batch triggered during last iteration is
+		still running, */
+		if (!buf_flush_start(buf_pool, BUF_FLUSH_LRU)) {
+			continue;
 		}
-	}
 
-	if (total_flushed) {
-		MONITOR_INC_VALUE_CUMULATIVE(
-			MONITOR_LRU_BATCH_TOTAL_PAGE,
-			MONITOR_LRU_BATCH_COUNT,
-			MONITOR_LRU_BATCH_PAGES,
-			total_flushed);
+		buf_flush_batch(buf_pool, BUF_FLUSH_LRU, scan_depth, 0, &n);
+
+		buf_flush_end(buf_pool, BUF_FLUSH_LRU);
+
+		buf_flush_common(BUF_FLUSH_LRU, n.flushed);
+
+		if (n.flushed) {
+			MONITOR_INC_VALUE_CUMULATIVE(
+				MONITOR_LRU_BATCH_FLUSH_TOTAL_PAGE,
+				MONITOR_LRU_BATCH_FLUSH_COUNT,
+				MONITOR_LRU_BATCH_FLUSH_PAGES,
+				n.flushed);
+		}
+
+		if (n.evicted) {
+			MONITOR_INC_VALUE_CUMULATIVE(
+				MONITOR_LRU_BATCH_EVICT_TOTAL_PAGE,
+				MONITOR_LRU_BATCH_EVICT_COUNT,
+				MONITOR_LRU_BATCH_EVICT_PAGES,
+				n.evicted);
+		}
+
+		total_flushed += (n.flushed + n.evicted);
 	}
 
 	return(total_flushed);
@@ -2382,18 +2277,28 @@ page_cleaner_sleep_if_needed(
 	ulint	next_loop_time)	/*!< in: time when next loop iteration
 				should start */
 {
+	/* No sleep if we are cleaning the buffer pool during the shutdown
+	with everything else finished */
+	if (srv_shutdown_state == SRV_SHUTDOWN_FLUSH_PHASE)
+		return;
+
 	ulint	cur_time = ut_time_ms();
 
 	if (next_loop_time > cur_time) {
 		/* Get sleep interval in micro seconds. We use
-		ut_min() to avoid long sleep in case of
-		wrap around. */
-		os_thread_sleep(ut_min(1000000,
-				(next_loop_time - cur_time)
-				 * 1000));
+		ut_min() to avoid long sleep in case of wrap around. */
+		ulint	sleep_us;
+
+		sleep_us = ut_min(1000000, (next_loop_time - cur_time) * 1000);
+
+		ib_int64_t	sig_count = os_event_reset(buf_flush_event);
+
+		os_event_wait_time_low(buf_flush_event, sleep_us, sig_count);
 	}
 }
 
+
+
 /******************************************************************//**
 page_cleaner thread tasked with flushing dirty pages from the buffer
 pools. As of now we'll have only one instance of this thread.
@@ -2421,7 +2326,6 @@ DECLARE_THREAD(buf_flush_page_cleaner_thread)(
 	fprintf(stderr, "InnoDB: page_cleaner thread running, id %lu\n",
 		os_thread_pf(os_thread_get_curr_id()));
 #endif /* UNIV_DEBUG_THREAD_CREATION */
-
 	buf_page_cleaner_is_active = TRUE;
 
 	while (srv_shutdown_state == SRV_SHUTDOWN_NONE) {
@@ -2434,12 +2338,12 @@ DECLARE_THREAD(buf_flush_page_cleaner_thread)(
 			last_activity = srv_get_activity_count();
 
 			/* Flush pages from flush_list if required */
-			page_cleaner_flush_pages_if_needed();
-			n_flushed = 0;
-		} else {
+			n_flushed += page_cleaner_flush_pages_if_needed();
+
+		} else if (srv_idle_flush_pct) {
 			n_flushed = page_cleaner_do_flush_batch(
-							PCT_IO(100),
-							LSN_MAX);
+				PCT_IO(100),
+				LSN_MAX);
 
 			if (n_flushed) {
 				MONITOR_INC_VALUE_CUMULATIVE(
@@ -2451,10 +2355,11 @@ DECLARE_THREAD(buf_flush_page_cleaner_thread)(
 		}
 
 		/* Flush pages from end of LRU if required */
-		n_flushed = buf_flush_LRU_tail();
+		buf_flush_LRU_tail();
 	}
 
 	ut_ad(srv_shutdown_state > 0);
+
 	if (srv_fast_shutdown == 2) {
 		/* In very fast shutdown we simulate a crash of
 		buffer pool. We are not required to do any flushing */
@@ -2621,9 +2526,11 @@ buf_flush_validate(
 
 	return(ret);
 }
+
 #endif /* UNIV_DEBUG || UNIV_BUF_DEBUG */
 #endif /* !UNIV_HOTBACKUP */
 
+
 #ifdef UNIV_DEBUG
 /******************************************************************//**
 Check if there are any dirty pages that belong to a space id in the flush
